# sage_setup: distribution = sagemath-groups
"""
Free Groups

Free groups and finitely presented groups are implemented as a wrapper
over the corresponding GAP objects.

A free group can be created by giving the number of generators, or their names.
It is also possible to create indexed generators::

    sage: G.<x,y,z> = FreeGroup();  G
    Free Group on generators {x, y, z}
    sage: FreeGroup(3)
    Free Group on generators {x0, x1, x2}
    sage: FreeGroup('a,b,c')
    Free Group on generators {a, b, c}
    sage: FreeGroup(3,'t')
    Free Group on generators {t0, t1, t2}

The elements can be created by operating with the generators, or by passing a list
with the indices of the letters to the group:

EXAMPLES::

    sage: G.<a,b,c> = FreeGroup()
    sage: a*b*c*a
    a*b*c*a
    sage: G([1,2,3,1])
    a*b*c*a
    sage: a * b / c * b^2
    a*b*c^-1*b^2
    sage: G([1,1,2,-1,-3,2])
    a^2*b*a^-1*c^-1*b

You can use call syntax to replace the generators with a set of
arbitrary ring elements::

    sage: g =  a * b / c * b^2
    sage: g(1,2,3)
    8/3
    sage: M1 = identity_matrix(2)
    sage: M2 = matrix([[1,1],[0,1]])
    sage: M3 = matrix([[0,1],[1,0]])
    sage: g([M1, M2, M3])
    [1 3]
    [1 2]

AUTHORS:

- Miguel Angel Marco Buzunariz
- Volker Braun
"""

# ****************************************************************************
#       Copyright (C) 2012 Miguel Angel Marco Buzunariz <mmarco@unizar.es>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.categories.groups import Groups
from sage.groups.group import Group
from sage.groups.libgap_wrapper import ParentLibGAP, ElementLibGAP
from sage.structure.unique_representation import CachedRepresentation
from sage.libs.gap.libgap import libgap
from sage.libs.gap.element import GapElement
from sage.rings.integer import Integer
from sage.rings.integer_ring import IntegerRing
from sage.misc.cachefunc import cached_method
from sage.misc.misc_c import prod
from sage.structure.sequence import Sequence
from sage.structure.element import coercion_model, parent
from sage.structure.richcmp import richcmp, richcmp_method


def is_FreeGroup(x):
    """
    Test whether ``x`` is a :class:`FreeGroup_class`.

    INPUT:

    - ``x`` -- anything

    OUTPUT: boolean

    EXAMPLES::

        sage: from sage.groups.free_group import is_FreeGroup
        sage: is_FreeGroup('a string')                                                  # needs sage.combinat
        False
        sage: is_FreeGroup(FreeGroup(0))
        True
        sage: is_FreeGroup(FreeGroup(index_set=ZZ))                                     # needs sage.combinat
        True
    """
    if isinstance(x, FreeGroup_class):
        return True
    from sage.groups.indexed_free_group import IndexedFreeGroup
    return isinstance(x, IndexedFreeGroup)


def _lexi_gen(zeroes=False):
    """
    Return a generator object that produces variable names suitable for the
    generators of a free group.

    INPUT:

    - ``zeroes`` -- boolean (default: ``False``); if ``True``, the
      integers appended to the output string begin at zero at the
      first iteration through the alphabet

    OUTPUT:

    Python generator object which outputs a character from the alphabet on each
    ``next()`` call in lexicographical order. The integer `i` is appended
    to the output string on the `i`-th iteration through the alphabet.

    EXAMPLES::

        sage: from sage.groups.free_group import _lexi_gen
        sage: itr = _lexi_gen()
        sage: F = FreeGroup([next(itr) for i in [1..10]]); F
        Free Group on generators {a, b, c, d, e, f, g, h, i, j}
        sage: it = _lexi_gen()
        sage: [next(it) for i in range(10)]
        ['a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j']
        sage: itt = _lexi_gen(True)
        sage: [next(itt) for i in range(10)]
        ['a0', 'b0', 'c0', 'd0', 'e0', 'f0', 'g0', 'h0', 'i0', 'j0']
        sage: test = _lexi_gen()
        sage: ls = [next(test) for i in range(3*26)]
        sage: ls[2*26:3*26]
        ['a2', 'b2', 'c2', 'd2', 'e2', 'f2', 'g2', 'h2', 'i2', 'j2', 'k2', 'l2', 'm2',
         'n2', 'o2', 'p2', 'q2', 'r2', 's2', 't2', 'u2', 'v2', 'w2', 'x2', 'y2', 'z2']

    TESTS::

        sage: from sage.groups.free_group import _lexi_gen
        sage: test = _lexi_gen()
        sage: ls = [next(test) for i in range(500)]
        sage: ls[234], ls[260]
        ('a9', 'a10')
    """
    count = Integer(0)
    while True:
        mwrap, ind = count.quo_rem(26)
        if mwrap == 0 and not zeroes:
            name = ''
        else:
            name = str(mwrap)
        name = chr(ord('a') + ind) + name
        yield name
        count += 1


class FreeGroupElement(ElementLibGAP):
    """
    A wrapper of GAP's Free Group elements.

    INPUT:

    - ``x`` -- something that determines the group element; either a
      :class:`~sage.libs.gap.element.GapElement` or the Tietze list
      (see :meth:`Tietze`) of the group element

    - ``parent`` -- the parent :class:`FreeGroup`

    EXAMPLES::

        sage: G = FreeGroup('a, b')
        sage: x = G([1, 2, -1, -2])
        sage: x
        a*b*a^-1*b^-1
        sage: y = G([2, 2, 2, 1, -2, -2, -2])
        sage: y
        b^3*a*b^-3
        sage: x*y
        a*b*a^-1*b^2*a*b^-3
        sage: y*x
        b^3*a*b^-3*a*b*a^-1*b^-1
        sage: x^(-1)
        b*a*b^-1*a^-1
        sage: x == x*y*y^(-1)
        True
    """

    def __init__(self, parent, x):
        """
        The Python constructor.

        See :class:`FreeGroupElement` for details.

        TESTS::

            sage: G.<a,b> = FreeGroup()
            sage: x = G([1, 2, -1, -1])
            sage: x # indirect doctest
            a*b*a^-2
            sage: y = G([2, 2, 2, 1, -2, -2, -1])
            sage: y # indirect doctest
            b^3*a*b^-2*a^-1

            sage: TestSuite(G).run()
            sage: TestSuite(x).run()
        """
        if not isinstance(x, GapElement):
            try:
                l = x.Tietze()
            except AttributeError:
                l = list(x)
            if len(l) > 0:
                if min(l) < -parent.ngens() or parent.ngens() < max(l):
                    raise ValueError('generators not in the group')
            if 0 in l:
                raise ValueError('zero does not denote a generator')
            i = 0
            while i < len(l)-1:
                if l[i] == -l[i+1]:
                    l.pop(i)
                    l.pop(i)
                    if i > 0:
                        i = i-1
                else:
                    i = i+1
            AbstractWordTietzeWord = libgap.eval('AbstractWordTietzeWord')
            x = AbstractWordTietzeWord(l, parent.gap().GeneratorsOfGroup())
        ElementLibGAP.__init__(self, parent, x)

    def __hash__(self):
        r"""
        TESTS::

            sage: G.<a,b> = FreeGroup()
            sage: hash(a*b*b*~a) == hash((1, 2, 2, -1))
            True
        """
        return hash(self.Tietze())

    def _latex_(self):
        r"""
        Return a LaTeX representation.

        OUTPUT: string; a valid LaTeX math command sequence

        EXAMPLES::

            sage: F.<a,b,c> = FreeGroup()
            sage: f = F([1, 2, 2, -3, -1]) * c^15 * a^(-23)
            sage: f._latex_()
            'a\\cdot b^{2}\\cdot c^{-1}\\cdot a^{-1}\\cdot c^{15}\\cdot a^{-23}'

            sage: F = FreeGroup(3)
            sage: f = F([1, 2, 2, -3, -1]) * F.gen(2)^11 * F.gen(0)^(-12)
            sage: f._latex_()
            'x_{0}\\cdot x_{1}^{2}\\cdot x_{2}^{-1}\\cdot x_{0}^{-1}\\cdot x_{2}^{11}\\cdot x_{0}^{-12}'

            sage: F.<a,b,c> = FreeGroup()
            sage: G = F /  (F([1, 2, 1, -3, 2, -1]), F([2, -1]))
            sage: f = G([1, 2, 2, -3, -1]) * G.gen(2)^15 * G.gen(0)^(-23)
            sage: f._latex_()
            'a\\cdot b^{2}\\cdot c^{-1}\\cdot a^{-1}\\cdot c^{15}\\cdot a^{-23}'

            sage: F = FreeGroup(4)
            sage: G = F.quotient((F([1, 2, 4, -3, 2, -1]), F([2, -1])))
            sage: f = G([1, 2, 2, -3, -1]) * G.gen(3)^11 * G.gen(0)^(-12)
            sage: f._latex_()
            'x_{0}\\cdot x_{1}^{2}\\cdot x_{2}^{-1}\\cdot x_{0}^{-1}\\cdot x_{3}^{11}\\cdot x_{0}^{-12}'
        """
        import re
        s = self._repr_()
        s = re.sub('([a-z]|[A-Z])([0-9]+)', r'\g<1>_{\g<2>}', s)
        s = re.sub(r'(\^)(-)([0-9]+)', r'\g<1>{\g<2>\g<3>}', s)
        s = re.sub(r'(\^)([0-9]+)', r'\g<1>{\g<2>}', s)
        s = s.replace('*', r'\cdot ')
        return s

    def __reduce__(self):
        """
        Implement pickling.

        TESTS::

            sage: F.<a,b> = FreeGroup()
            sage: a.__reduce__()
            (Free Group on generators {a, b}, ((1,),))
            sage: (a*b*a^-1).__reduce__()
            (Free Group on generators {a, b}, ((1, 2, -1),))
        """
        return (self.parent(), (self.Tietze(),))

    @cached_method
    def Tietze(self):
        """
        Return the Tietze list of the element.

        The Tietze list of a word is a list of integers that represent
        the letters in the word.  A positive integer `i` represents
        the letter corresponding to the `i`-th generator of the group.
        Negative integers represent the inverses of generators.

        OUTPUT: tuple of integers

        EXAMPLES::

            sage: G.<a,b> = FreeGroup()
            sage: a.Tietze()
            (1,)
            sage: x = a^2 * b^(-3) * a^(-2)
            sage: x.Tietze()
            (1, 1, -2, -2, -2, -1, -1)

        TESTS::

            sage: type(a.Tietze())
            <... 'tuple'>
            sage: type(a.Tietze()[0])
            <class 'sage.rings.integer.Integer'>
        """
        tl = self.gap().TietzeWordAbstractWord()
        return tuple(tl.sage())

    def fox_derivative(self, gen, im_gens=None, ring=None):
        r"""
        Return the Fox derivative of ``self`` with respect to a given
        generator ``gen`` of the free group.

        Let `F` be a free group with free generators
        `x_1, x_2, \ldots, x_n`. Let `j \in \left\{ 1, 2, \ldots ,
        n \right\}`. Let `a_1, a_2, \ldots, a_n` be `n`
        invertible elements of a ring `A`. Let `a : F \to A^\times`
        be the (unique) homomorphism from `F` to the multiplicative
        group of invertible elements of `A` which sends each `x_i`
        to `a_i`. Then, we can define a map `\partial_j : F \to A`
        by the requirements that

        .. MATH::

            \partial_j (x_i) = \delta_{i, j}
            \qquad \qquad \text{ for all indices } i \text{ and } j

        and

        .. MATH::

            \partial_j (uv) = \partial_j(u) + a(u) \partial_j(v)
            \qquad \qquad \text{ for all } u, v \in F .

        This map `\partial_j` is called the `j`-th Fox derivative
        on `F` induced by `(a_1, a_2, \ldots, a_n)`.

        The most well-known case is when `A` is the group ring
        `\ZZ [F]` of `F` over `\ZZ`, and when `a_i = x_i \in A`.
        In this case, `\partial_j` is simply called the `j`-th
        Fox derivative on `F`.

        INPUT:

        - ``gen`` -- the generator with respect to which the
          derivative will be computed. If this is `x_j`, then the
          method will return `\partial_j`.

        - ``im_gens`` -- (optional) the images of the generators
          (given as a list or iterable). This is the list
          `(a_1, a_2, \ldots, a_n)`.
          If not provided, it defaults to
          `(x_1, x_2, \ldots, x_n)` in the group ring
          `\ZZ [F]`.

        - ``ring`` -- (optional) the ring in which the elements
          of the list  `(a_1, a_2, \ldots, a_n)` lie. If not
          provided, this ring is inferred from these elements.

        OUTPUT:

        The fox derivative of ``self`` with respect to ``gen``
        (induced by ``im_gens``).
        By default, it is an element of the group algebra with
        integer coefficients.
        If ``im_gens`` are provided, the result lives in the
        algebra where ``im_gens`` live.

        EXAMPLES::

            sage: G = FreeGroup(5)
            sage: G.inject_variables()
            Defining x0, x1, x2, x3, x4
            sage: (~x0*x1*x0*x2*~x0).fox_derivative(x0)
            -x0^-1 + x0^-1*x1 - x0^-1*x1*x0*x2*x0^-1
            sage: (~x0*x1*x0*x2*~x0).fox_derivative(x1)
            x0^-1
            sage: (~x0*x1*x0*x2*~x0).fox_derivative(x2)
            x0^-1*x1*x0
            sage: (~x0*x1*x0*x2*~x0).fox_derivative(x3)
            0

        If ``im_gens`` is given, the images of the generators are
        mapped to them::

            sage: F = FreeGroup(3)
            sage: a = F([2,1,3,-1,2])
            sage: a.fox_derivative(F([1]))
            x1 - x1*x0*x2*x0^-1
            sage: R.<t> = LaurentPolynomialRing(ZZ)
            sage: a.fox_derivative(F([1]),[t,t,t])
            t - t^2
            sage: S.<t1,t2,t3> = LaurentPolynomialRing(ZZ)
            sage: a.fox_derivative(F([1]),[t1,t2,t3])
            -t2*t3 + t2
            sage: R.<x,y,z> = QQ[]
            sage: a.fox_derivative(F([1]),[x,y,z])
            -y*z + y
            sage: a.inverse().fox_derivative(F([1]),[x,y,z])
            (z - 1)/(y*z)

        The optional parameter ``ring`` determines the ring `A`::

            sage: u = a.fox_derivative(F([1]), [1,2,3], ring=QQ)
            sage: u
            -4
            sage: parent(u)
            Rational Field
            sage: u = a.fox_derivative(F([1]), [1,2,3], ring=R)
            sage: u
            -4
            sage: parent(u)
            Multivariate Polynomial Ring in x, y, z over Rational Field

        TESTS::

            sage: F = FreeGroup(3)
            sage: a = F([])
            sage: a.fox_derivative(F([1]))
            0
            sage: R.<t> = LaurentPolynomialRing(ZZ)
            sage: a.fox_derivative(F([1]),[t,t,t])
            0
        """
        if gen not in self.parent().generators():
            raise ValueError("Fox derivative can only be computed with respect to generators of the group")
        l = list(self.Tietze())
        if im_gens is None:
            F = self.parent()
            R = F.algebra(IntegerRing())
            R_basis = R.basis()
            symb = [R_basis[a] for a in F.gens()]
            symb += reversed([R_basis[a.inverse()] for a in F.gens()])
            if ring is not None:
                R = ring
                symb = [R(i) for i in symb]
        else:
            if ring is None:
                R = Sequence(im_gens).universe()
            else:
                R = ring
            symb = list(im_gens)
            symb += reversed([a**(-1) for a in im_gens])
        i = gen.Tietze()[0]
        # so gen is the i-th generator of the free group

        a = R.zero()
        coef = R.one()
        while l:
            b = l.pop(0)
            if b == i:
                a += coef * R.one()
                coef *= symb[b-1]
            elif b == -i:
                a -= coef * symb[b]
                coef *= symb[b]
            elif b > 0:
                coef *= symb[b-1]
            else:
                coef *= symb[b]
        return a

    @cached_method
    def syllables(self):
        r"""
        Return the syllables of the word.

        Consider a free group element `g = x_1^{n_1} x_2^{n_2} \cdots
        x_k^{n_k}`. The uniquely-determined subwords `x_i^{e_i}`
        consisting only of powers of a single generator are called the
        syllables of `g`.

        OUTPUT:

        The tuple of syllables. Each syllable is given as a pair
        `(x_i, e_i)` consisting of a generator and a nonzero integer.

        EXAMPLES::

            sage: G.<a,b> = FreeGroup()
            sage: w = a^2 * b^-1 * a^3
            sage: w.syllables()
            ((a, 2), (b, -1), (a, 3))
        """
        g = self.gap().UnderlyingElement()
        k = g.NumberSyllables().sage()
        exponent_syllable = libgap.eval('ExponentSyllable')
        generator_syllable = libgap.eval('GeneratorSyllable')
        result = []
        gen = self.parent().gen
        for i in range(k):
            exponent = exponent_syllable(g, i+1).sage()
            generator = gen(generator_syllable(g, i+1).sage() - 1)
            result.append((generator, exponent))
        return tuple(result)

    def __call__(self, *values):
        """
        Replace the generators of the free group by corresponding
        elements of the iterable ``values`` in the group element
        ``self``.

        INPUT:

        - ``*values`` -- a sequence of values, or a
          list/tuple/iterable of the same length as the number of
          generators of the free group

        OUTPUT: the product of ``values`` in the order and with exponents
        specified by ``self``

        EXAMPLES::

            sage: G.<a,b> = FreeGroup()
            sage: w = a^2 * b^-1 * a^3
            sage: w(1, 2)
            1/2
            sage: w(2, 1)
            32
            sage: w.subs(b=1, a=2)   # indirect doctest
            32

        TESTS::

            sage: w([1, 2])
            1/2
            sage: w((1, 2))
            1/2
            sage: w(i+1 for i in range(2))
            1/2

        Check that :issue:`25017` is fixed::

            sage: F = FreeGroup(2)
            sage: x0, x1 = F.gens()
            sage: u = F(1)
            sage: parent(u.subs({x1:x0})) is F
            True

            sage: F = FreeGroup(2)
            sage: x0, x1 = F.gens()
            sage: u = x0*x1
            sage: u.subs({x0:3, x1:2})
            6
            sage: u.subs({x0:1r, x1:2r})
            2
            sage: M0 = matrix(ZZ,2,[1,1,0,1])
            sage: M1 = matrix(ZZ,2,[1,0,1,1])
            sage: u.subs({x0: M0, x1: M1})
            [2 1]
            [1 1]

        TESTS::

            sage: F.<x,y> = FreeGroup()
            sage: F.one().subs(x=x, y=1)
            Traceback (most recent call last):
            ...
            TypeError: no common canonical parent for objects with parents: 'Free Group on generators {x, y}' and 'Integer Ring'
        """
        if len(values) == 1:
            try:
                values = list(values[0])
            except TypeError:
                pass
        G = self.parent()
        if len(values) != G.ngens():
            raise ValueError('number of values has to match the number of generators')
        replace = dict(zip(G.gens(), values))
        new_parent = coercion_model.common_parent(*[parent(v) for v in values])
        try:
            return new_parent.prod(replace[gen] ** power
                                   for gen, power in self.syllables())
        except AttributeError:
            return prod(new_parent(replace[gen]) ** power
                        for gen, power in self.syllables())


def FreeGroup(n=None, names='x', index_set=None, abelian=False, **kwds):
    """
    Construct a Free Group.

    INPUT:

    - ``n`` -- integer (default: ``None``); the number of
      generators (if not specified the ``names`` are counted)

    - ``names`` -- string or list/tuple/iterable of strings (default:
      ``'x'``); the generator names or name prefix

    - ``index_set`` -- (optional) an index set for the generators; if
      specified then the optional keyword ``abelian`` can be used

    - ``abelian`` -- boolean (default: ``False``); whether to construct a free
      abelian group or a free group

    .. NOTE::

        If you want to create a free group, it is currently preferential to
        use ``Groups().free(...)`` as that does not load GAP.

    EXAMPLES::

        sage: G.<a,b> = FreeGroup();  G
        Free Group on generators {a, b}
        sage: H = FreeGroup('a, b')
        sage: G is H
        True
        sage: FreeGroup(0)
        Free Group on generators {}

    The entry can be either a string with the names of the generators,
    or the number of generators and the prefix of the names to be
    given. The default prefix is ``'x'`` ::

        sage: FreeGroup(3)
        Free Group on generators {x0, x1, x2}
        sage: FreeGroup(3, 'g')
        Free Group on generators {g0, g1, g2}
        sage: FreeGroup()
        Free Group on generators {x}

    We give two examples using the ``index_set`` option::

        sage: FreeGroup(index_set=ZZ)                                                   # needs sage.combinat
        Free group indexed by Integer Ring
        sage: FreeGroup(index_set=ZZ, abelian=True)                                     # needs sage.combinat
        Free abelian group indexed by Integer Ring

    TESTS::

        sage: G1 = FreeGroup(2, 'a,b')
        sage: G2 = FreeGroup('a,b')
        sage: G3.<a,b> = FreeGroup()
        sage: G1 is G2, G2 is G3
        (True, True)
    """
    # Support Freegroup('a,b') syntax
    if n is not None:
        try:
            n = Integer(n)
        except TypeError:
            names = n
            n = None
    # derive n from counting names
    if n is None:
        if isinstance(names, str):
            n = len(names.split(','))
        else:
            names = list(names)
            n = len(names)
    from sage.structure.category_object import normalize_names
    names = normalize_names(n, names)
    if index_set is not None or abelian:
        if abelian:
            from sage.groups.indexed_free_group import IndexedFreeAbelianGroup
            return IndexedFreeAbelianGroup(index_set, names=names, **kwds)

        from sage.groups.indexed_free_group import IndexedFreeGroup
        return IndexedFreeGroup(index_set, names=names, **kwds)
    return FreeGroup_class(names, **kwds)


@richcmp_method
class FreeGroup_class(CachedRepresentation, Group, ParentLibGAP):
    """
    A class that wraps GAP's FreeGroup.

    See :func:`FreeGroup` for details.

    TESTS::

        sage: G = FreeGroup('a, b')
        sage: TestSuite(G).run()
        sage: G.category()
        Category of infinite groups
    """
    Element = FreeGroupElement

    def __init__(self, generator_names, gap_group=None):
        """
        Python constructor.

        INPUT:

        - ``generator_names`` -- tuple of strings; the names of the
          generators

        - ``libgap_free_group`` -- a LibGAP free group (default: ``None``);
          the LibGAP free group to wrap (if ``None``, a suitable group will be
          constructed)

        TESTS::

            sage: G.<a,b> = FreeGroup() # indirect doctest
            sage: G
            Free Group on generators {a, b}
            sage: G.variable_names()
            ('a', 'b')
        """
        if gap_group is None:
            gap_group = libgap.FreeGroup(generator_names)
        ParentLibGAP.__init__(self, gap_group)
        if not generator_names:
            cat = Groups().Finite()
        else:
            cat = Groups().Infinite()
        Group.__init__(self, category=cat)
        self._gen_names = generator_names
        try:
            self._assign_names(generator_names)
        except ValueError:
            pass

    def __hash__(self):
        """
        Make hashable.

        EXAMPLES::

            sage: F = FreeGroup(3)
            sage: F.__hash__() == hash(F._gen_names)
            True
        """
        return hash(self._gen_names)

    def __richcmp__(self, other, op):
        """
        Rich comparison of ``self`` and ``other``.

        EXAMPLES::

            sage: G1 = FreeGroup('a, b')
            sage: gg = libgap.FreeGroup('x', 'y')
            sage: G2 = FreeGroup('a, b', gap_group=gg)
            sage: G1 == G2
            True
            sage: G1 is G2
            False
            sage: G3 = FreeGroup('x, y')
            sage: G1 == G3
            False
            sage: G2 == G3
            False
        """
        if not isinstance(other, self.__class__):
            from sage.structure.richcmp import op_NE
            return (op == op_NE)
        return richcmp(self._gen_names, other._gen_names, op)

    def _repr_(self):
        """
        TESTS::

            sage: G = FreeGroup('a, b')
            sage: G # indirect doctest
            Free Group on generators {a, b}
            sage: G._repr_()
            'Free Group on generators {a, b}'
        """
        return 'Free Group on generators {' + ', '.join(self._gen_names) + '}'

    def rank(self):
        """
        Return the number of generators of ``self``.

        Alias for :meth:`ngens`.

        OUTPUT: integer

        EXAMPLES::

            sage: G = FreeGroup('a, b');  G
            Free Group on generators {a, b}
            sage: G.rank()
            2
            sage: H = FreeGroup(3, 'x')
            sage: H
            Free Group on generators {x0, x1, x2}
            sage: H.rank()
            3
        """
        return self.ngens()

    def _gap_init_(self):
        """
        Return the string used to construct the object in gap.

        EXAMPLES::

            sage: G = FreeGroup(3)
            sage: G._gap_init_()
            'FreeGroup(["x0", "x1", "x2"])'
        """
        gap_names = ['"' + s + '"' for s in self._gen_names]
        gen_str = ', '.join(gap_names)
        return 'FreeGroup(['+gen_str+'])'

    def _element_constructor_(self, *args, **kwds):
        """
        TESTS::

            sage: G.<a,b> = FreeGroup()
            sage: G([1, 2, 1]) # indirect doctest
            a*b*a
            sage: G([1, 2, -2, 1, 1, -2]) # indirect doctest
            a^3*b^-1

            sage: G( a.gap() )
            a
            sage: type(_)
            <class 'sage.groups.free_group.FreeGroup_class_with_category.element_class'>

        Check that conversion between free groups follow the convention that
        names are preserved::

            sage: F = FreeGroup('a,b')
            sage: G = FreeGroup('b,a')
            sage: G(F.gen(0))
            a
            sage: F(G.gen(0))
            b
            sage: a,b = F.gens()
            sage: G(a^2*b^-3*a^-1)
            a^2*b^-3*a^-1

        Check that :issue:`17246` is fixed::

            sage: F = FreeGroup(0)
            sage: F([])
            1

        Check that 0 isn't considered the identity::

            sage: F = FreeGroup('x')
            sage: F(0)
            Traceback (most recent call last):
            ...
            TypeError: 'sage.rings.integer.Integer' object is not iterable
        """
        if len(args) != 1:
            return self.element_class(self, *args, **kwds)
        x = args[0]
        if x == 1 or x == [] or x == ():
            return self.one()
        try:
            P = x.parent()
        except AttributeError:
            return self.element_class(self, x, **kwds)
        if isinstance(P, FreeGroup_class):
            names = {P._gen_names[abs(i)-1] for i in x.Tietze()}
            if names.issubset(self._gen_names):
                return self([i.sign()*(self._gen_names.index(P._gen_names[abs(i)-1])+1)
                             for i in x.Tietze()])
            else:
                raise ValueError('generators of %s not in the group' % x)
        return self.element_class(self, x, **kwds)

    def abelian_invariants(self):
        r"""
        Return the Abelian invariants of ``self``.

        The Abelian invariants are given by a list of integers
        `i_1 \dots i_j`, such that the abelianization of the
        group is isomorphic to

        .. MATH::

            \ZZ / (i_1) \times \dots \times \ZZ / (i_j)

        EXAMPLES::

            sage: F.<a,b> = FreeGroup()
            sage: F.abelian_invariants()
            (0, 0)
        """
        return (0,) * self.ngens()

    def quotient(self, relations, **kwds):
        """
        Return the quotient of ``self`` by the normal subgroup generated
        by the given elements.

        This quotient is a finitely presented groups with the same
        generators as ``self``, and relations given by the elements of
        ``relations``.

        INPUT:

        - ``relations`` -- list/tuple/iterable with the elements of
          the free group
        - further named arguments, that are passed to the constructor
          of a finitely presented group

        OUTPUT:

        A finitely presented group, with generators corresponding to
        the generators of the free group, and relations corresponding
        to the elements in ``relations``.

        EXAMPLES::

            sage: F.<a,b> = FreeGroup()
            sage: F.quotient([a*b^2*a, b^3])
            Finitely presented group < a, b | a*b^2*a, b^3 >

        Division is shorthand for :meth:`quotient` ::

            sage: F /  [a*b^2*a, b^3]
            Finitely presented group < a, b | a*b^2*a, b^3 >

        Relations are converted to the free group, even if they are not
        elements of it (if possible) ::

            sage: F1.<a,b,c,d> = FreeGroup()
            sage: F2.<a,b> = FreeGroup()
            sage: r = a*b/a
            sage: r.parent()
            Free Group on generators {a, b}
            sage: F1/[r]
            Finitely presented group < a, b, c, d | a*b*a^-1 >
        """
        from sage.groups.finitely_presented import FinitelyPresentedGroup
        return FinitelyPresentedGroup(self,
                                      tuple(map(self, relations)), **kwds)

    __truediv__ = quotient



# sage_setup: distribution = sagemath-combinat
# sage.doctest: needs sage.combinat
"""
Indexed Free Groups

Free groups and free abelian groups implemented using an indexed set of
generators.

AUTHORS:

- Travis Scrimshaw (2013-10-16): Initial version
"""

##############################################################################
#       Copyright (C) 2013 Travis Scrimshaw <tscrim at ucdavis.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
##############################################################################

from sage.categories.groups import Groups
from sage.categories.poor_man_map import PoorManMap
from sage.groups.group import Group, AbelianGroup
from sage.monoids.indexed_free_monoid import (IndexedMonoid,
                                              IndexedFreeMonoidElement,
                                              IndexedFreeAbelianMonoidElement)
from sage.misc.cachefunc import cached_method
import sage.data_structures.blas_dict as blas
from sage.rings.integer import Integer
from sage.rings.infinity import infinity
from sage.sets.family import Family


class IndexedGroup(IndexedMonoid):
    """
    Base class for free (abelian) groups whose generators are indexed
    by a set.

    TESTS:

    We check finite properties::

        sage: G = Groups().free(index_set=ZZ)
        sage: G.is_finite()
        False
        sage: G = Groups().free(index_set='abc')
        sage: G.is_finite()
        False
        sage: G = Groups().free(index_set=[])
        sage: G.is_finite()
        True

    ::

        sage: G = Groups().Commutative().free(index_set=ZZ)
        sage: G.is_finite()
        False
        sage: G = Groups().Commutative().free(index_set='abc')
        sage: G.is_finite()
        False
        sage: G = Groups().Commutative().free(index_set=[])
        sage: G.is_finite()
        True
    """
    def order(self):
        r"""
        Return the number of elements of ``self``, which is `\infty` unless
        this is the trivial group.

        EXAMPLES::

            sage: G = Groups().free(index_set=ZZ)
            sage: G.order()
            +Infinity
            sage: G = Groups().Commutative().free(index_set='abc')
            sage: G.order()
            +Infinity
            sage: G = Groups().Commutative().free(index_set=[])
            sage: G.order()
            1
        """
        return self.cardinality()

    def rank(self):
        """
        Return the rank of ``self``.

        This is the number of generators of ``self``.

        EXAMPLES::

            sage: G = Groups().free(index_set=ZZ)
            sage: G.rank()
            +Infinity
            sage: G = Groups().free(index_set='abc')
            sage: G.rank()
            3
            sage: G = Groups().free(index_set=[])
            sage: G.rank()
            0

        ::

            sage: G = Groups().Commutative().free(index_set=ZZ)
            sage: G.rank()
            +Infinity
            sage: G = Groups().Commutative().free(index_set='abc')
            sage: G.rank()
            3
            sage: G = Groups().Commutative().free(index_set=[])
            sage: G.rank()
            0
        """
        return self.group_generators().cardinality()

    @cached_method
    def group_generators(self):
        """
        Return the group generators of ``self``.

        EXAMPLES::

            sage: G = Groups.free(index_set=ZZ)
            sage: G.group_generators()
            Lazy family (Generator map from Integer Ring to
             Free group indexed by Integer Ring(i))_{i in Integer Ring}
            sage: G = Groups().free(index_set='abcde')
            sage: sorted(G.group_generators())
            [F['a'], F['b'], F['c'], F['d'], F['e']]
        """
        if self._indices.cardinality() == infinity:
            gen = PoorManMap(self.gen, domain=self._indices, codomain=self, name="Generator map")
            return Family(self._indices, gen)
        return Family(self._indices, self.gen)

    gens = group_generators


class IndexedFreeGroup(IndexedGroup, Group):
    """
    An indexed free group.

    EXAMPLES::

        sage: G = Groups().free(index_set=ZZ)
        sage: G
        Free group indexed by Integer Ring
        sage: G = Groups().free(index_set='abcde')
        sage: G
        Free group indexed by {'a', 'b', 'c', 'd', 'e'}
    """
    def __init__(self, indices, prefix, category=None, **kwds):
        """
        Initialize ``self``.

        TESTS::

            sage: G = Groups().free(index_set=ZZ)
            sage: TestSuite(G).run()
            sage: G = Groups().free(index_set='abc')
            sage: TestSuite(G).run()
        """
        category = Groups().or_subcategory(category)
        IndexedGroup.__init__(self, indices, prefix, category, **kwds)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        TESTS::

            sage: Groups().free(index_set=ZZ)       # indirect doctest
            Free group indexed by Integer Ring
        """
        return 'Free group indexed by {}'.format(self._indices)

    @cached_method
    def one(self):
        """
        Return the identity element of ``self``.

        EXAMPLES::

            sage: G = Groups().free(ZZ)
            sage: G.one()
            1
        """
        return self.element_class(self, ())

    def gen(self, x):
        """
        The generator indexed by ``x`` of ``self``.

        EXAMPLES::

            sage: G = Groups().free(index_set=ZZ)
            sage: G.gen(0)
            F[0]
            sage: G.gen(2)
            F[2]
        """
        if x not in self._indices:
            raise IndexError("{} is not in the index set".format(x))
        try:
            return self.element_class(self, ((self._indices(x),1),))
        except TypeError: # Backup (if it is a string)
            return self.element_class(self, ((x,1),))

    class Element(IndexedFreeMonoidElement):
        def __len__(self):
            """
            Return the length of ``self``.

            EXAMPLES::

                sage: G = Groups().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: elt = a*c^-3*b^-2*a
                sage: elt.length()
                7
                sage: len(elt)
                7

                sage: G = Groups().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: elt = a*c^-3*b^-2*a
                sage: elt.length()
                7
                sage: len(elt)
                7
            """
            return sum(abs(exp) for gen,exp in self._monomial)

        length = __len__

        def _mul_(self, other):
            """
            Multiply ``self`` by ``other``.

            EXAMPLES::

                sage: G = Groups().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: a*b^2*e*d
                F[0]*F[1]^2*F[4]*F[3]
                sage: (a*b^2*d^2) * (d^-4*b*e)
                F[0]*F[1]^2*F[3]^-2*F[1]*F[4]
                sage: (a*b^-2*d^2) * (d^-2*b^2*a^-1)
                1
            """
            if not self._monomial:
                return other
            if not other._monomial:
                return self

            ret = list(self._monomial)
            rhs = list(other._monomial)
            while ret and rhs and ret[-1][0] == rhs[0][0]:
                rhs[0] = (rhs[0][0], rhs[0][1] + ret.pop()[1])
                if rhs[0][1] == 0:
                    rhs.pop(0)
            ret += rhs
            return self.__class__(self.parent(), tuple(ret))

        def __invert__(self):
            """
            Return the inverse of ``self``.

            EXAMPLES::

                sage: G = Groups().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: x = a*b^2*e^-1*d; ~x
                F[3]^-1*F[4]*F[1]^-2*F[0]^-1
                sage: x * ~x
                1
            """
            return self.__class__(self.parent(),
                   tuple((x[0], -x[1]) for x in reversed(self._monomial)))

        def to_word_list(self):
            """
            Return ``self`` as a word represented as a list whose entries
            are the pairs ``(i, s)`` where ``i`` is the index and ``s`` is
            the sign.

            EXAMPLES::

                sage: G = Groups().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: x = a*b^2*e*a^-1
                sage: x.to_word_list()
                [(0, 1), (1, 1), (1, 1), (4, 1), (0, -1)]
            """
            sign = lambda x: 1 if x > 0 else -1 # It is never 0
            return [ (k, sign(e)) for k,e in self._sorted_items()
                     for dummy in range(abs(e))]


class IndexedFreeAbelianGroup(IndexedGroup, AbelianGroup):
    """
    An indexed free abelian group.

    EXAMPLES::

        sage: G = Groups().Commutative().free(index_set=ZZ)
        sage: G
        Free abelian group indexed by Integer Ring
        sage: G = Groups().Commutative().free(index_set='abcde')
        sage: G
        Free abelian group indexed by {'a', 'b', 'c', 'd', 'e'}
    """
    def __init__(self, indices, prefix, category=None, **kwds):
        """
        Initialize ``self``.

        TESTS::

            sage: G = Groups().Commutative().free(index_set=ZZ)
            sage: TestSuite(G).run()
            sage: G = Groups().Commutative().free(index_set='abc')
            sage: TestSuite(G).run()
        """
        category = Groups().or_subcategory(category)
        IndexedGroup.__init__(self, indices, prefix, category, **kwds)

    def _repr_(self):
        """
        TESTS::

            sage: Groups.Commutative().free(index_set=ZZ)
            Free abelian group indexed by Integer Ring
        """
        return 'Free abelian group indexed by {}'.format(self._indices)

    def _element_constructor_(self, x=None):
        """
        Create an element of ``self`` from ``x``.

        EXAMPLES::

            sage: G = Groups().Commutative().free(index_set=ZZ)
            sage: G(G.gen(2))
            F[2]
            sage: G([[1, 3], [-2, 12]])
            F[-2]^12*F[1]^3
            sage: G({1: 3, -2: 12})
            F[-2]^12*F[1]^3
            sage: G(-5)
            F[-5]

        TESTS::

            sage: G([(1, 3), (1, -5)])
            F[1]^-2

            sage: G([(42, 0)])
            1
            sage: G([(42, 3), (42, -3)])
            1
            sage: G({42: 0})
            1
        """
        if isinstance(x, (list, tuple)):
            d = {}
            for k, v in x:
                if k in d:
                    d[k] += v
                else:
                    d[k] = v
            x = d
        if isinstance(x, dict):
            x = {k: v for k, v in x.items() if v != 0}
        return IndexedGroup._element_constructor_(self, x)

    @cached_method
    def one(self):
        """
        Return the identity element of ``self``.

        EXAMPLES::

            sage: G = Groups().Commutative().free(index_set=ZZ)
            sage: G.one()
            1
        """
        return self.element_class(self, {})

    def gen(self, x):
        """
        The generator indexed by ``x`` of ``self``.

        EXAMPLES::

            sage: G = Groups().Commutative().free(index_set=ZZ)
            sage: G.gen(0)
            F[0]
            sage: G.gen(2)
            F[2]
        """
        if x not in self._indices:
            raise IndexError("{} is not in the index set".format(x))
        try:
            return self.element_class(self, {self._indices(x):1})
        except TypeError: # Backup (if it is a string)
            return self.element_class(self, {x:1})

    class Element(IndexedFreeAbelianMonoidElement, IndexedFreeGroup.Element):
        def _mul_(self, other):
            """
            Multiply ``self`` by ``other``.

            EXAMPLES::

                sage: G = Groups().Commutative().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: a*b^2*e^-1*d
                F[0]*F[1]^2*F[3]*F[4]^-1
                sage: (a*b^2*d^2) * (d^-4*b^-2*e)
                F[0]*F[3]^-2*F[4]
                sage: (a*b^-2*d^2) * (d^-2*b^2*a^-1)
                1
            """
            return self.__class__(self.parent(),
                                  blas.add(self._monomial, other._monomial))

        def __invert__(self):
            """
            Return the inverse of ``self``.

            EXAMPLES::

                sage: G = Groups().Commutative().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: x = a*b^2*e^-1*d; ~x
                F[0]^-1*F[1]^-2*F[3]^-1*F[4]
                sage: x * ~x
                1
            """
            return self ** -1

        def __floordiv__(self, a):
            """
            Return the division of ``self`` by ``a``.

            EXAMPLES::

                sage: G = Groups().Commutative().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: elt = a*b*c^3*d^2; elt
                F[0]*F[1]*F[2]^3*F[3]^2
                sage: elt // a
                F[1]*F[2]^3*F[3]^2
                sage: elt // c
                F[0]*F[1]*F[2]^2*F[3]^2
                sage: elt // (a*b*d^2)
                F[2]^3
                sage: elt // a^4
                F[0]^-3*F[1]*F[2]^3*F[3]^2
            """
            return self * ~a

        def __pow__(self, n):
            """
            Raise ``self`` to the power of ``n``.

            EXAMPLES::

                sage: G = Groups().Commutative().free(index_set=ZZ)
                sage: a,b,c,d,e = [G.gen(i) for i in range(5)]
                sage: x = a*b^2*e^-1*d; x
                F[0]*F[1]^2*F[3]*F[4]^-1
                sage: x^3
                F[0]^3*F[1]^6*F[3]^3*F[4]^-3
                sage: x^0
                1
                sage: x^-3
                F[0]^-3*F[1]^-6*F[3]^-3*F[4]^3
            """
            if not isinstance(n, (int, Integer)):
                raise TypeError("Argument n (= {}) must be an integer".format(n))
            if n == 1:
                return self
            if n == 0:
                return self.parent().one()
            return self.__class__(self.parent(), {k:v*n for k,v in self._monomial.items()})



# sage_setup: distribution = sagemath-groups
# sage.doctest: needs sage.rings.number_field
r"""
Cactus Groups

AUTHORS:

- Travis Scrimshaw (1-2023): initial version
"""

# ****************************************************************************
#       Copyright (C) 2023 Travis Scrimshaw <tcscrims at gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.misc.cachefunc import cached_method
from sage.misc.lazy_attribute import lazy_attribute
from sage.structure.element import MultiplicativeGroupElement, parent
from sage.structure.unique_representation import UniqueRepresentation
from sage.structure.richcmp import richcmp
from sage.matrix.constructor import matrix
from sage.categories.groups import Groups
from sage.groups.group import Group
from sage.groups.kernel_subgroup import KernelSubgroup
from sage.combinat.permutation import Permutations
from sage.sets.family import Family
from sage.graphs.graph import Graph
from itertools import combinations


class CactusGroup(UniqueRepresentation, Group):
    r"""
    The cactus group.

    The `n`-fruit cactus group `J_n` is the group generated by `s_{pq}`
    for `1 \leq p < q \leq n` with relations:

    - `s_{pq}^2 = 1`
    - `s_{pq} s_{kl} = s_{kl} s_{pq}` if the intervals `[p, q]` and `[k, l]`
      are disjoint, and
    - `s_{pq} s_{kl} = s_{p+q-l,p+q-k} s_{pq}` if `[k, l] \subseteq [p, q]`.

    INPUT:

    - ``n`` -- integer

    EXAMPLES:

    We construct the cactus group `J_3` and do some basic computations::

        sage: J3 = groups.misc.Cactus(3)
        sage: s12,s13,s23 = J3.group_generators()
        sage: s12 * s13
        s[1,2]*s[1,3]
        sage: x = s12 * s23; x
        s[1,2]*s[2,3]
        sage: x^4
        s[1,2]*s[2,3]*s[1,2]*s[2,3]*s[1,2]*s[2,3]*s[1,2]*s[2,3]
        sage: s12 * s13 == s13 * s23
        True

    We verify the key equality in Lemma 2.3 in [White2015]_, which shows
    that `J_5` is generated by `s_{1q}`::

        sage: J5 = groups.misc.Cactus(5)
        sage: gens = J5.group_generators()
        sage: all(gens[(p,q)] == gens[(1,q)] * gens[(1,q-p+1)] * gens[(1,q)]
        ....:     for p in range(1, 6) for q in range(p+1, 6))
        True
    """
    def __init__(self, n):
        r"""
        Initialize ``self``.

        TESTS::

            sage: J3 = groups.misc.Cactus(3)
            sage: it = iter(J3)
            sage: elts = [next(it) for _ in range(100)]
            sage: TestSuite(J3).run(elements=elts[::7], skip='_test_enumerated_set_contains')

        We run this test separately because the words grow very long, very
        quickly. This means the code needs to check a lot of normalizations,
        so we limit the number of tests::

            sage: J3._test_enumerated_set_contains(max_runs=500)

        ::

            sage: J4 = groups.misc.Cactus(4)
            sage: it = iter(J4)
            sage: elts = [next(it) for _ in range(100)]
            sage: TestSuite(J4).run(elements=elts[::8])
        """
        self._n = n
        ell = len(str(n))
        names = ['s{}{}'.format('0' * (ell - len(str(i))) + str(i),
                                '0' * (ell - len(str(j))) + str(j))
                 for i in range(1, self._n + 1)
                 for j in range(i + 1, self._n + 1)]
        cat = Groups().FinitelyGeneratedAsMagma()
        if n > 2:
            cat = cat.Infinite()
        else:
            cat = cat.Finite()
        Group.__init__(self, category=cat)
        # Group.__init__ doesn't take a "names" parameter
        self._assign_names(names)

    @lazy_attribute
    def _WG(self):
        r"""
        Get the graph for the right-angled Coxeter group that ``self``
        embeds into (set theoretically) and set the ``_subsets`` and
        ``_subsets_inv`` attributes.

        We do this initialization lazily in order to make the creation of
        the parent quick. However, this is used to normalize the product
        of elements.

        EXAMPLES::

            sage: J4 = groups.misc.Cactus(4)
            sage: J4._WG
            Graph on 11 vertices
            sage: J4._subsets
            [frozenset({1, 2}), frozenset({1, 3}), frozenset({1, 4}), frozenset({2, 3}),
             frozenset({2, 4}), frozenset({3, 4}), frozenset({1, 2, 3}), frozenset({1, 2, 4}),
             frozenset({1, 3, 4}), frozenset({2, 3, 4}), frozenset({1, 2, 3, 4})]
        """
        n = self._n
        I = list(range(1, n+1))
        PS = sum(([frozenset(A) for A in combinations(I, k)] for k in range(2,n+1)), [])
        G = Graph([list(range(len(PS))),
                   [[i,j,-1] for j in range(1, len(PS)) for i in range(j)
                    if PS[i] & PS[j] not in [frozenset(), PS[i], PS[j]]]
                   ], format='vertices_and_edges')
        self._subsets = PS
        self._subsets_inv = {X: i for i,X in enumerate(PS)}
        return G

    def right_angled_coxeter_group(self):
        r"""
        Return the right-angled Coxeter group that ``self``
        (set-theoretically) embeds into.

        This is defined following [Most2019]_, where it was called the
        diagram group. It has generators (of order `2`) indexed by subsets
        of `\{1, \ldots, n\}` that commute if and only if one subset is
        contained in the other or they are disjoint. For the pure cactus
        group, this is also a group homomorphism, otherwise it is a
        group 1-cocycle [BCL2022]_.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J3.right_angled_coxeter_group()
            Coxeter group over Rational Field with Coxeter matrix:
            [ 1 -1 -1  2]
            [-1  1 -1  2]
            [-1 -1  1  2]
            [ 2  2  2  1]
        """
        from sage.rings.rational_field import QQ
        from sage.combinat.root_system.coxeter_group import CoxeterGroup
        from sage.combinat.root_system.coxeter_matrix import CoxeterMatrix
        return CoxeterGroup(CoxeterMatrix(self._WG), base_ring=QQ)

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: groups.misc.Cactus(3)
            Cactus Group with 3 fruit
        """
        return "Cactus Group with {} fruit".format(self._n)

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: latex(J3)
            J_{3}
        """
        return "J_{{{}}}".format(self._n)

    def n(self):
        """
        Return the value `n`.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J3.n()
            3
        """
        return self._n

    @cached_method
    def group_generators(self):
        """
        Return the group generators of ``self``.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J3.group_generators()
            Finite family {(1, 2): s[1,2], (1, 3): s[1,3], (2, 3): s[2,3]}
        """
        l = [(i, j) for i in range(1, self._n + 1)
             for j in range(i + 1, self._n + 1)]
        return Family(l, lambda x: self.element_class(self, [x]))

    @cached_method
    def gens(self) -> tuple:
        """
        Return the generators of ``self`` as a tuple.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J3.gens()
            (s[1,2], s[1,3], s[2,3])
        """
        return tuple(self.group_generators())

    def gen(self, i, j=None):
        r"""
        Return the `i`-th generator of ``self`` or the generator `s_{ij}`.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J3.gen(1)
            s[1,3]
            sage: J3.gen(1,2)
            s[1,2]
            sage: J3.gen(0,2)
            Traceback (most recent call last):
            ...
            ValueError: s[0,2] is not a valid generator
            sage: J3.gen(1,4)
            Traceback (most recent call last):
            ...
            ValueError: s[1,4] is not a valid generator
            sage: J3.gen(2,1)
            Traceback (most recent call last):
            ...
            ValueError: s[2,1] is not a valid generator
        """
        if j is None:
            return self.gens()[i]
        if not (1 <= i < j <= self._n):
            raise ValueError(f"s[{i},{j}] is not a valid generator")
        return self.element_class(self, [(i,j)])

    @cached_method
    def one(self):
        r"""
        Return the identity element in ``self``.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J3.one()
            1
        """
        return self.element_class(self, [])

    def _coerce_map_from_(self, G):
        r"""
        Return if there is a coerce map from ``G``.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J5 = groups.misc.Cactus(5)
            sage: PJ3 = groups.misc.PureCactus(3)
            sage: PJ5 = groups.misc.PureCactus(5)

            sage: J3._coerce_map_from_(J5)
            False
            sage: J5._coerce_map_from_(J3)
            True
            sage: J3._coerce_map_from_(PJ3)
            True
            sage: J3._coerce_map_from_(PJ5)
            False
            sage: J5._coerce_map_from_(PJ3)
            True
            sage: J5._coerce_map_from_(PJ5)
            True
        """
        if isinstance(G, CactusGroup):
            return G._n <= self._n
        elif isinstance(G, PureCactusGroup):
            return G.n() <= self._n
        return super()._coerce_map_from_(G)

    def _element_constructor_(self, x):
        r"""
        Construct an element of ``self`` from ``x``.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J5 = groups.misc.Cactus(5)
            sage: PJ3 = groups.misc.PureCactus(3)

            sage: J5(J3.an_element())
            s[1,2]*s[2,3]*s[1,3]
            sage: J3(J5.an_element())
            s[1,2]*s[2,3]*s[1,3]
            sage: it = iter(PJ3)
            sage: [J3(next(it)) for _ in range(3)]
            [1, s[1,2]*s[2,3]*s[1,2]*s[1,3], s[2,3]*s[1,2]*s[2,3]*s[1,3]]

            sage: J3([[1,2], [1,3], [2,3]])
            s[1,3]
        """
        if parent(x) is self:
            return x
        if isinstance(x, PureCactusGroup.Element):
            x = x.value
        if isinstance(x, CactusGroup.Element):
            if any(d[1] > self._n for d in x._data):
                raise ValueError(f"{x} is not an element of {self}")
            return self.element_class(self, x._data)
        ret = self.element_class(self, x)
        ret._normalize()
        return ret

    def _an_element_(self):
        r"""
        Return an element of ``self``.

        TESTS::

            sage: J1 = groups.misc.Cactus(1)
            sage: J1._an_element_()
            1
            sage: J2 = groups.misc.Cactus(2)
            sage: J2._an_element_()
            s[1,2]
            sage: J3 = groups.misc.Cactus(3)
            sage: x = J3._an_element_(); x
            s[1,2]*s[2,3]*s[1,3]

            sage: x._normalize()
            sage: x
            s[1,2]*s[2,3]*s[1,3]
        """
        if self._n <= 1:
            return self.one()
        if self._n == 2:
            return self.element_class(self, [(1, 2)])
        return self.element_class(self, [(1, 2), (2, 3), (1, 3)])

    def random_element(self, max_length=10):
        r"""
        Return a random element of ``self`` of length at most ``max_length``.

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: J3.random_element()  # random
            s[1,2]*s[2,3]*s[1,2]*s[1,3]
        """
        from sage.misc.prandom import randint
        l = randint(0, max_length)
        gens = list(self.group_generators())
        ret = self.one()
        for _ in range(l):
            ret *= gens[randint(0, len(gens) - 1)]
        return ret

    def bilinear_form(self, t=None):
        r"""
        Return the ``t``-bilinear form of ``self``.

        We define a bilinear form `B` on the group algebra by

        .. MATH::

            B(s_{ij}, s_{pq}) = \begin{cases}
            1 & \text{if } i = p, j = q, \\
            -t & \text{if } [i, j] \not\subseteq [p, q] \text{ and }
            [p, q] \not\subseteq [i, j], \\
            0 & \text{otherwise}.
            \end{cases}

        In other words, it is `1` if `s_{ij} = s_{pq}`, `-t` if `s_{ij}`
        and `s_{pq}` generate a free group, and `0` otherwise (they commute
        or almost commute).

        INPUT:

        - ``t`` -- (default: `t` in `\ZZ[t]`) the variable `t`

        EXAMPLES::

            sage: J = groups.misc.Cactus(4)
            sage: B = J.bilinear_form()
            sage: B
            [ 1  0  0 -t -t  0]
            [ 0  1  0  0 -t -t]
            [ 0  0  1  0  0  0]
            [-t  0  0  1  0 -t]
            [-t -t  0  0  1  0]
            [ 0 -t  0 -t  0  1]

        We reorder the generators so the bilinear form is more
        "Coxeter-like". In particular, when we remove the generator
        `s_{1,4}`, we recover the bilinear form in Example 6.2.5
        of [DJS2003]_::

            sage: J.gens()
            (s[1,2], s[1,3], s[1,4], s[2,3], s[2,4], s[3,4])
            sage: S = SymmetricGroup(6)
            sage: g = S([1,4,6,2,5,3])
            sage: B.permute_rows_and_columns(g, g)
            sage: B
            [ 1 -t  0  0 -t  0]
            [-t  1 -t  0  0  0]
            [ 0 -t  1 -t  0  0]
            [ 0  0 -t  1 -t  0]
            [-t  0  0 -t  1  0]
            [ 0  0  0  0  0  1]
        """
        if t is None:
            from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing
            from sage.rings.integer_ring import ZZ
            t = PolynomialRing(ZZ, 't').gen()
        R = t.parent()
        ret = []
        K = self.group_generators().keys()
        for x in K:
            for y in K:
                if x is y:
                    ret.append(R.one())
                elif (x[1] < y[0] or x[0] > y[1] or  # Disjoint
                      (x[0] <= y[0] and y[1] <= x[1]) or  # y <= x
                      (y[0] <= x[0] and x[1] <= y[1])):  # x <= y
                    ret.append(R.zero())
                else:
                    ret.append(-t)
        return matrix(R, len(K), len(K), ret)

    def geometric_representation_generators(self, t=None):
        r"""
        Return the matrices corresponding to the generators of ``self``.

        We construct a representation over `R = \ZZ[t]` of `J_n` as follows.
        Let `E` be the vector space over `R` spanned by `\{\epsilon_v\}_v`,
        where `v` is a generator of `J_n`. Fix some generator `v`, and
        let `E_v` denote the span of `\epsilon_u - \epsilon_{u'}`,
        where `u'` is the reflected interval of `u` in `v`, over all
        `u` such that `u \subset v`. Let `F_v` denote the orthogonal
        complement of `R \epsilon_v \oplus E_v` with respect to the
        :meth:`bilinear form <bilinear_form>` `B`. We define the action
        of `v` on `E` by

        .. MATH::

            \rho(v) = -I |_{R\epsilon_v \oplus E_v} \oplus I |_{F_v}.

        By Theorem 6.2.3 of [DJS2003]_, this defines a representation of `J_n`.
        It is expected that this is a faithful representation (see
        Remark 6.2.4 of [DJS2003]_). As this arises from a blow-up and an
        analog of the geometric representation of the corresponding
        Coxeter group (the symmetric group), we call this the
        *geometric representation*.

        INPUT:

        - ``t`` -- (default: `t` in `\ZZ[t]`) the variable `t`

        EXAMPLES::

            sage: J3 = groups.misc.Cactus(3)
            sage: list(J3.geometric_representation_generators())
            [
            [ -1   0 2*t]  [ 0  0  1]  [  1   0   0]
            [  0   1   0]  [ 0 -1  0]  [  0   1   0]
            [  0   0   1], [ 1  0  0], [2*t   0  -1]
            ]

        We ran the following code with ``max_tests = 15000`` and did
        not find a counterexample to the faithfulness of this representation::

            sage: visited = set([J3.one()])
            sage: cur = set([(J3.one(), J3.one().to_matrix())])
            sage: mats = set([J3.one().to_matrix()])
            sage: RG = list(J3.geometric_representation_generators())
            sage: count = 0
            sage: max_tests = 1000
            sage: while cur:
            ....:     count += 1
            ....:     if count >= max_tests:
            ....:         break
            ....:     elt, mat = cur.pop()
            ....:     for g,r in zip(J3, RG):
            ....:         val = elt * g
            ....:         if val in visited:
            ....:             continue
            ....:         visited.add(val)
            ....:         matp = mat * r
            ....:         matp.set_immutable()
            ....:         assert matp not in mats, f"not injective {val} \n{matp}"
            ....:         mats.add(matp)
            ....:         cur.add((val, matp))
        """
        B = self.bilinear_form(t)
        F = B.base_ring().fraction_field()
        K = self.group_generators().keys()
        from sage.modules.free_module import FreeModule
        V = FreeModule(F, len(K))
        basis = V.basis()
        ret = {}
        for ik, k in enumerate(K):
            E = [basis[ik]]
            ME = matrix(E)
            # The only non-trivial elements are those reflected by the interval k
            for low in range(k[0], k[1] + 1):
                for high in range(low + 1, k[1] + 1):
                    v = (low, high)
                    vp = (k[0] + k[1] - high, k[0] + k[1] - low)
                    if v == vp:
                        continue
                    elt = basis[K.index(v)] - basis[K.index(vp)]
                    if elt not in ME.row_space():
                        E.append(elt)
                        ME = ME.stack(elt)
            # Get the orthogonal complement wrt to the bilinear form B
            Fv = (ME * B).right_kernel().basis_matrix()
            T = ME.stack(Fv).transpose()
            rho = matrix.diagonal(F, [-1] * len(E) + [1] * (len(K) - len(E)))
            ret[k] = T * rho * ~T
        return Family(K, lambda k: ret[k])

    class Element(MultiplicativeGroupElement):
        """
        An element of a cactus group.
        """
        def __init__(self, parent, data):
            """
            Initialize ``self``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: elt = J3.an_element()
                sage: TestSuite(elt).run()
            """
            self._data = tuple(data)
            MultiplicativeGroupElement.__init__(self, parent)

        def _repr_(self):
            """
            Return a string representation of ``self``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: J3.one()
                1
                sage: J3.an_element()
                s[1,2]*s[2,3]*s[1,3]
            """
            if not self._data:
                return '1'
            return '*'.join(f"s[{i},{j}]" for i,j in self._data)

        def _latex_(self):
            """
            Return a latex representation of ``self``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: latex(J3.one())
                1
                sage: latex(J3.an_element())
                s_{1,2} s_{2,3} s_{1,3}
            """
            if not self._data:
                return '1'
            return " ".join(f"s_{{{i},{j}}}" for i,j in self._data)

        def _unicode_art_(self):
            """
            Return unicode art of ``self``.

            EXAMPLES::

                sage: J4 = groups.misc.Cactus(4)
                sage: s12,s13,s14,s23,s24,s34 = J4.gens()
                sage: unicode_art(s12 * s23 * s13)
                s₁₂ s₂₃ s₁₃
                sage: unicode_art(J4.one())
                1
                sage: J12 = groups.misc.Cactus(12)
                sage: unicode_art(J12.gen(1,3))
                s₁,₃
                sage: unicode_art(J12.gen(3,11))
                s₃,₁₁
            """
            from sage.typeset.unicode_art import unicode_subscript, unicode_art
            if not self._data:
                return unicode_art('1')
            if self.parent()._n < 10:
                return unicode_art(' '.join('s{}{}'.format(unicode_subscript(p), unicode_subscript(q)) for p,q in self._data))
            return unicode_art(' '.join('s{},{}'.format(unicode_subscript(p), unicode_subscript(q)) for p,q in self._data))

        def __hash__(self):
            r"""
            Return the hash of ``self``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: elt = J3.gen(1,2) * J3.gen(2,3) * J3.gen(1,3)
                sage: hash(elt) == hash((((1,2), (2,3), (1,3))))
                True
            """
            return hash(self._data)

        def _richcmp_(self, other, op):
            r"""
            Compare ``self`` and ``other``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: elt = J3.an_element()
                sage: elt == J3.one()
                False
                sage: elt != J3.one()
                True
                sage: s12,s13,s23 = J3.gens()
                sage: elt == s12*s23*s13
                True
            """
            return richcmp(self._data, other._data, op)

        def _mul_(self, other):
            r"""
            Return the product of ``self`` and ``other``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: s12,s13,s23 = J3.gens()
                sage: s12*s23
                s[1,2]*s[2,3]
                sage: s12*s13
                s[1,2]*s[1,3]
                sage: s13*s12
                s[2,3]*s[1,3]
                sage: J3.one() * (s13*s12*s13*s12*s23*s13)
                s[2,3]*s[1,2]*s[2,3]*s[1,3]
            """
            ret = type(self)(self.parent(), self._data + other._data)
            ret._normalize()
            return ret

        def __invert__(self):
            r"""
            Return the inverse of ``self``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: s12,s13,s23 = J3.gens()
                sage: elt = s12*s23*s13
                sage: ~elt
                s[1,2]*s[2,3]*s[1,3]
                sage: elt * elt
                1
            """
            ret = type(self)(self.parent(), reversed(self._data))
            ret._normalize()
            return ret

        def to_permutation(self):
            r"""
            Return the image of ``self`` under the canonical projection
            to the permutation group.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: s12,s13,s23 = J3.gens()
                sage: s12.to_permutation()
                [2, 1, 3]
                sage: s23.to_permutation()
                [1, 3, 2]
                sage: s13.to_permutation()
                [3, 2, 1]
                sage: elt = s12*s23*s13
                sage: elt.to_permutation()
                [1, 3, 2]

                sage: J7 = groups.misc.Cactus(7)
                sage: J7.group_generators()[3,6].to_permutation()
                [1, 2, 6, 5, 4, 3, 7]

            We check that this respects the multiplication order
            of permutations::

                sage: P3 = Permutations(3)
                sage: elt = s12*s23
                sage: elt.to_permutation() == P3(s12) * P3(s23)
                True
                sage: Permutations.options.mult='r2l'
                sage: elt.to_permutation() == P3(s12) * P3(s23)
                True
                sage: Permutations.options.mult='l2r'
            """
            n = self.parent().n()
            P = Permutations(n)
            ret = P.one()
            for x in self._data:
                lst = list(range(1, n + 1))
                lst[x[0] - 1:x[1]] = list(reversed(lst[x[0] - 1:x[1]]))
                ret *= P(lst)
            return ret

        def _matrix_(self):
            r"""
            Return ``self`` as a matrix.

            The resulting matrix is the :meth:`geometric representation
            <sage.groups.cactus_group.CactusGroup.geometric_representation_generators>`
            of ``self``.

            EXAMPLES::

                sage: J3 = groups.misc.Cactus(3)
                sage: s12,s13,s23 = J3.gens()
                sage: s12.to_matrix()
                [ -1   0 2*t]
                [  0   1   0]
                [  0   0   1]
                sage: (s12*s13).to_matrix()
                [2*t   0  -1]
                [  0  -1   0]
                [  1   0   0]
                sage: (s13*s23).to_matrix()
                [2*t   0  -1]
                [  0  -1   0]
                [  1   0   0]
                sage: (s13*s12).to_matrix()
                [  0   0   1]
                [  0  -1   0]
                [ -1   0 2*t]
                sage: all(x.to_matrix() * y.to_matrix() == (x*y).to_matrix()
                ....:     for x in J3.gens() for y in J3.gens())
                True
            """
            G = self.parent().geometric_representation_generators()
            ret = G[(1,2)].parent().one()
            for x in self._data:
                ret *= G[x]
            ret.set_immutable()
            return ret

        to_matrix = _matrix_

        def _normalize(self):
            r"""
            Return the word for ``self`` in normalized form.

            ALGORITHM:

            We perform the normalization by using the lexicographically
            minimum reduced word for the corresponding right-angled Coxeter
            group (RACG) element under the (set-theoretic) embedding
            introduced by [Most2019]_. This embedding is a group 1-cocycle
            and also realizes the cactus group as a subgroup of `W \rtimes S_n`,
            where `W` is the RACG (see also [Yu2022]_). See Section 2
            of [BCL2022]_ for precise statements.

            TESTS::

                sage: J6 = groups.misc.Cactus(6)
                sage: s26 = J6.gen(2,6)
                sage: s45 = J6.gen(4,5)
                sage: s13 = J6.gen(1,3)
                sage: s26 * s45 * s13 == s26 * s45 * s13  # indirect doctest
                True

                sage: J4 = groups.misc.Cactus(4)
                sage: s12,s13,s14,s23,s24,s34 = J4.gens()
                sage: s12 * (s12 * s23)
                s[2,3]
                sage: (s12 * s23) * s23
                s[1,2]
                sage: s23 * s13 * s34
                s[2,3]*s[1,3]*s[3,4]
            """
            P = self.parent()
            n = P._n
            G = P._WG  # The defining graph

            # Convert to an element in the right-angled Coxeter group
            perm = list(range(1,n+1))
            word = []
            for p,q in self._data:
                word.append(P._subsets_inv[frozenset(perm[p-1:q])])
                perm[p-1:q] = reversed(perm[p-1:q])

            # Normalize the word
            # This code works for any right-angled Coxeter group
            pos = 0
            supp = sorted(set(word))
            while pos < len(word) - 1:
                cur = word[pos]
                for i in supp:
                    if i > cur:
                        break
                    if G.has_edge(cur, i):
                        continue
                    did_swap = False
                    for j in range(pos+1, len(word)):
                        if word[j] == i:
                            word.pop(j)
                            if cur == i:  # canceling s_i s_i = 1
                                word.pop(pos)
                                pos = -1  # start again at the beginning
                                break
                            word.insert(pos, i)
                            did_swap = True
                            break
                        if G.has_edge(i, word[j]):
                            break
                    if did_swap:
                        break
                pos += 1

            # Convert back
            ret = []
            perm = list(range(1,n+1))
            for i in word:
                X = P._subsets[i]
                pos = [j for j,val in enumerate(perm) if val in X]
                for j in range(len(pos)//2):
                    perm[pos[j]], perm[pos[-j-1]] = perm[pos[-j-1]], perm[pos[j]]
                pos.sort()
                assert all(pos[k] + 1 == pos[k+1] for k in range(len(pos)-1))
                ret.append((pos[0]+1, pos[-1]+1))

            self._data = tuple(ret)


class PureCactusGroup(KernelSubgroup):
    r"""
    The pure cactus group.

    The *pure cactus group* `PJ_n` is the kernel of the natural
    surjection of the cactus group `J_n` onto the symmetric group
    `S_n`. In particular, we have the following (non-split) exact sequence:

    .. MATH::

        1 \longrightarrow PJ_n \longrightarrow J_n \longrightarrow S_n
        \longrightarrow 1.
    """
    def __init__(self, n):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)
            sage: it = iter(PJ3)
            sage: elts = [next(it) for _ in range(10)]
            sage: TestSuite(PJ3).run(elements=elts)
        """
        J = CactusGroup(n)
        from sage.groups.perm_gps.permgroup_named import SymmetricGroup
        S = SymmetricGroup(n)
        KernelSubgroup.__init__(self, S.coerce_map_from(J))

    def _repr_(self):
        """
        Return a string representation of ``self``.

        EXAMPLES::

            sage: groups.misc.PureCactus(3)
            Pure Cactus Group with 3 fruit
        """
        return "Pure Cactus Group with {} fruit".format(self.n())

    def _latex_(self):
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)
            sage: latex(PJ3)
            PJ_{3}
        """
        return "PJ_{{{}}}".format(self.n())

    @cached_method
    def n(self):
        """
        Return the value `n`.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)
            sage: PJ3.n()
            3
        """
        return self.ambient().n()

    def gen(self, i):
        r"""
        Return the ``i``-th generator of ``self``.

        EXAMPLES::

            sage: PJ3 = groups.misc.PureCactus(3)
            sage: PJ3.gen(0)
            s[2,3]*s[1,2]*s[2,3]*s[1,3]
            sage: PJ3.gen(1)
            s[1,2]*s[2,3]*s[1,2]*s[1,3]
            sage: PJ3.gen(5)
            Traceback (most recent call last):
            ...
            IndexError: tuple index out of range
        """
        return self.gens()[i]

    @cached_method
    def gens(self) -> tuple:
        r"""
        Return the generators of ``self``.

        ALGORITHM:

        We use :wikipedia:`Schreier's_lemma` and compute the traversal
        using the lex minimum elements (defined by the order of the
        generators of the ambient cactus group).

        EXAMPLES:

        We verify Corollary A.2 of [BCL2022]_::

            sage: PJ3 = groups.misc.PureCactus(3)
            sage: PJ3.gens()
            (s[2,3]*s[1,2]*s[2,3]*s[1,3], s[1,2]*s[2,3]*s[1,2]*s[1,3])
            sage: a, b = PJ3.gens()
            sage: a * b  # they are inverses of each other
            1

            sage: J3 = groups.misc.Cactus(3)
            sage: gen = (J3.gen(1,2)*J3.gen(1,3))^3
            sage: gen
            s[1,2]*s[2,3]*s[1,2]*s[1,3]
            sage: gen == b
            True
        """
        from sage.arith.misc import factorial
        J = self.ambient()
        G = J.gens()
        one = J.one()
        n = self.n()
        nfac = factorial(n)
        reprs = {one.to_permutation(): one}
        next_level = [one]
        while len(reprs) < nfac:
            cur = next_level
            next_level = []
            for val in cur:
                for g in G:
                    temp = val * g
                    p = temp.to_permutation()
                    if p in reprs:
                        continue
                    reprs[p] = temp
                    next_level.append(temp)
        gens = []
        for s in reprs.values():
            for g in G:
                val = s * g * ~(reprs[(s*g).to_permutation()])
                if val == one or val in gens:
                    continue
                gens.append(val)
        return tuple([self(g) for g in gens])



# sage_setup: distribution = sagemath-groups
r"""
Type ``groups.lie.<tab>`` to access examples of Lie groups.
"""
from sage.misc.lazy_import import lazy_import as _lazy_import

# We use lazy import because the module depends on sage.manifolds and thus on sage.symbolic
_lazy_import('sage.groups.lie_gps.nilpotent_lie_group', 'NilpotentLieGroup', as_='Nilpotent')
